;;; moirai-dired-side.el -*- lexical-binding: t; -*-

(defcustom dired-side-buffer-ignore-list 'ivy-ignore-buffers
  "What list of regexp your completion engine uses to exclude buffers ?
Needed to prevent displaying the same side buffer in two places at the same time.")

;; window-preserved-size

(defcustom dired-side-window-parameters
  '((slot . 0) (side . left) (select . t)
    (preserve-size (t . t)) (window-width . 35)
    (window-parameters . ((no-other-window . t)
                          (no-delete-other-windows . t))))
  "Alist of parameters to apply when creating a dired side window.")

(defun dired-buffer-p (buffer &optional _action)
  "Is this BUFFER using `dired-mode'?"
  (eq 'dired-mode
      (buffer-local-value 'major-mode (window-normalize-buffer buffer))))

(defun dired-side-window-p (window)
  "Is this WINDOW using `dired' in a side-window?"
  (and (dired-buffer-p (window-buffer window))
       (window-parameter window 'window-side)))

(defun display-buffer--in-dired-side-window (buffer alist)
  "Display BUFFER in a side window with `dired-side-window-parameters' as ALIST.

The side parameter that may be inherited from a selected dired side window.
Only apply for buffers using `dired-mode' without `inhibit-same-window'."
  (when (and (dired-buffer-p buffer)
             (not (cdr (assq 'inhibit-same-window alist))))
    (if (dired-side-window-p (selected-window))
        (let* ((default-side (assoc 'side dired-side-window-parameters))
               (side (window-parameter (selected-window) 'window-side))
               (side-parameters (cl-subst `(side . ,side) default-side
                                          dired-side-window-parameters)))
          (display-buffer-in-side-window buffer side-parameters))
      (display-buffer-in-side-window buffer dired-side-window-parameters))))

(let ((old-side-buffers))
  (defun dired-side-window-update (frame)
    "Update on FRAME the local settings of last and current buffers
targeted by `global-side-window-mode'."
    (unless (minibuffer-window-active-p (selected-window))
      ;; register the active dired side windows
      (let (current-side-buffers)
        (walk-windows
         (lambda (window) (when (dired-side-window-p window)
                            (push (buffer-name (window-buffer window))
                                  current-side-buffers)))
         'nominibuf frame)

        ;; deactivate in the buffers not displayed anymore
        ;; and clean killed buffers from the old-side-buffers list
        (dolist (buffer (cl-nset-difference old-side-buffers
                                            current-side-buffers))
          (if (buffer-live-p (get-buffer buffer))
              (with-current-buffer (get-buffer buffer)
                (kill-local-variable 'display-buffer-overriding-action)
                (dired-hide-details-mode 0)))
          (setq old-side-buffers (cl-delete buffer old-side-buffers))
          (set dired-side-buffer-ignore-list
               (delete (concat "\\`" buffer "\\'") ivy-ignore-buffers)))

        ;; activate in the new buffers
        (dolist (buffer (cl-nset-difference current-side-buffers
                                            old-side-buffers))
          (with-current-buffer buffer
            (add-to-list dired-side-buffer-ignore-list
                         (concat "\\`" buffer "\\'"))
            (dired-hide-details-mode 1)
            (setq-local display-buffer-overriding-action
                        '(display-buffer--in-dired-side-window))))

        ;; save the currents side window buffers
        (setq old-side-buffers current-side-buffers)))))

(define-minor-mode global-dired-side-window-mode
  "Global minor mode that trigger `dired-side-window-update' when a buffer
take place in a window to set locally the following display settings :

  - dired buffers called from a side window will display again on the side window,
  - except calls specifying the `inhibit-same-window' window parameter,
that will be free to obey to to the defaults mechanisms,
eg : `display-buffer-alist' and `display-buffer-base-actions'."
  :init-value nil
  :global t
  :lighter ""
  (if global-dired-side-window-mode
      (add-hook 'window-buffer-change-functions #'dired-side-window-update)
    (remove-hook 'window-buffer-change-functions #'dired-side-window-update)))

(defmacro dired-side-window--with-windows (&rest body)
  "Eval BODY within a side window with `dired-mode' if that exists.
Returns t if at least one has been found."
  (declare (indent 0) (debug t))
  (let ((gen (gensym)))
    `(let (found)
       (walk-windows
        (lambda (,gen)
          (with-selected-window ,gen
            (when (dired-side-window-p ,gen)
              (progn ,@body)
              (setq found t)))))
       found)))

;;;###autoload
(defun dired-side-window-toggle (&optional arg)
  "Open or close a side window with a dired buffer inside at `default-directory',
With an \\[universal-argument], call for a project directory instead."
  (interactive "P")
  (unless (and (null arg)
               (dired-side-window--with-windows (window--delete (selected-window))))
    (let ((display-buffer-overriding-action '(display-buffer--in-dired-side-window)))
      (dired (if arg (project-root (project-current t)) ".")))))

(provide 'moirai-dired-side)

(global-set-key (kbd "s-d") #'dired-side-window-toggle)
(global-dired-side-window-mode 1)
 ;; (dired-side-window-global-mode 0)

;; recommended settings ?
;; switch-to-buffer-obey-display-actions nil (default)
;; ivy-use-ignore-default 'always
